import unittest
from exos import when

class TestWhen(unittest.TestCase):
    def test_eager(self):
        a = 42
        c = when(
            a < 10, 'less than 10',
            a == 42, 'the answer!',
            'something else'
        )
        self.assertEqual(c, 'the answer!')

    def test_lazy(self):
        class Stateful:
            def __init__(self, value):
                self.value = value
            def inc(self):
                self.value += 1
                return self.value

        state = Stateful(40)
        result = when(
            lambda: state.inc() == 42, 'first try!',
            lambda: state.inc() == 42, 'second try!',
            lambda: state.inc() == 42, 'third try!',
            lambda: state.inc(), 'needed more tries'
        )

        self.assertEqual(result, 'second try!')
        self.assertEqual(state.value, 42)



import unittest
from exos import extend

class TestExtend(unittest.TestCase):
    def test_shallow(self):
        self.assertEqual(
            extend({'a': 'A', 'b': 'B'}, {'c': 'C'}),
            {'a': 'A', 'b': 'B', 'c': 'C'}
        )

    def test_deep(self):
        d1 = {
            'package': 'exos',
            'motto': 'expressions over statements',
            'versions': {
                'beta': 'nightly'
            }
        }
        d2 = {
            'versions': {
                'beta': 'NIGHTLY'
            }
        }

        d = extend(d1, d2)
        self.assertTrue('package' in d)
        self.assertTrue(d['versions']['beta'] == 'NIGHTLY')

    def test_immutability(self):
        d = {'a': 42}
        _ = extend(d, d, d, d, d, {'b': 100})
        self.assertEqual(d, {'a': 42})



"""
decorators module
"""

__author__ = 'Bruno Lange'
__email__ = 'blangeram@gmail.com'
__license__ = 'MIT'

from functools import partial, wraps
from inspect import getfullargspec
from collections import defaultdict
from .utils import hashabledict, NotInCache

def curry(fn):
    """
    Decorator for currying functions.
    >>> @curry
    ... def volume(a, b, c):
    ...     return a*b*c
    ...
    >>> volume(1,2,3) == volume(1)(2)(3) == volume(1,2)(3) == volume(1)(2,3) == 6
    """
    _args = getfullargspec(fn).args

    @wraps(fn)
    def curried(*args, **kwargs):
        return (
            fn(*args, **kwargs) if len(_args) - len(args) == 0 else
            curry(partial(fn, *args, **kwargs))
        )
    return curried


def memoize(fn):
    """
    Decorator that provides automatic caching for referentially transparent
    functions.
    """
    cache = fn.cache = defaultdict(lambda: defaultdict(NotInCache))
    @wraps(fn)
    def memoized(*args, **kwargs):
        key0, key1 = str(args), hashabledict(kwargs) if kwargs else ''
        value = cache[key0][key1]
        if isinstance(value, NotInCache):
            value = cache[key0][key1] = fn(*args, **kwargs)
        return value
    return memoized



"""
utility functions and classes
"""

from functools import reduce

__author__ = 'Bruno Lange'
__email__ = 'blangeram@gmail.com'
__license__ = 'MIT'

def pairs(*args):
    def fold(acc, curr):
        if len(acc[-1]) < 2:
            acc[-1].append(curr)
        else:
            acc.append([curr])
        return acc
    return reduce(fold, args, [[]])

class hashabledict(dict):
    """
    A hashable dicitionary that comes with the cost of immutability.
    Any attempts to modify it after its inception yields a RunTimeError
    """
    def __readonly__(self, *args, **kwargs):
        raise RuntimeError('Cannot modify hashabledict')

    __setitem__ = __readonly__
    __delitem__ = __readonly__
    pop = __readonly__
    popitem = __readonly__
    clear = __readonly__
    update = __readonly__
    setdefault = __readonly__
    del __readonly__

    def __key(self):
        return tuple((k, self[k]) for k in sorted(self))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()

class NotInCache(object):
    pass



""" exos.py
EXpressions Over Statements: extended functional tools in Python.
"""

from functools import partial, reduce, wraps
from operator import iconcat
from inspect import getfullargspec
from collections import defaultdict
from .io import each, ueach, print_each
from .decorators import curry, memoize
from .exceptions import NonExhaustivePattern
from .utils import pairs

__author__ = 'Bruno Lange'
__email__ = 'blangeram@gmail.com'
__license__ = 'MIT'
__all__ = [
    'each',
    'ueach',
    'print_each',
    'curry',
    'memoize',
    'when',
    'flip',
    'map_attr',
    'map_method',
    'flatten',
    'zip_with_map',
    'extend',
    'reduce_right',
    'compose'
]

def match(subject, *args):
    for test, value in pairs(*args):
        if test is subject:
            return value
    raise NonExhaustivePattern

def when(*args):
    """
    The declarative version of a switch statement.
    >>> a = 42
    >>> c = when(
        a < 4,   'less than 4',
        a < 10,  'less than 10',
        a == 42, 'the answer!',
    )
    >>> print(c)
    The answer!

    If you'd like to defer evaluation of either the predicate
    or the actual value, use a lambda or a partial constructor
    to emulate laziness.
    """
    _pairs = pairs(*args)
    last = _pairs[-1]
    if len(last) < 2:
        last.insert(0, True)

    for predicate, value in _pairs:
        predicate = predicate() if callable(predicate) else predicate
        if predicate:
            return value() if callable(value) else value
    raise NonExhaustivePattern()

def flip(fn):
    spec = getfullargspec(fn)
    arity = len(spec.args) - len(spec.defaults or ())
    if arity < 2:
        return fn
    def flipped(*args, **kwargs):
        swapped = (args[1], args[0]) + args[2:]
        return (
            fn(*swapped, **kwargs) if len(args) == arity else
            partial(fn, *swapped, **kwargs)
        )
    return flipped

def map_attr(attr):
    """
    Returns a mapper function for attribute extraction

    map_attr('auth_user') <=> lambda account_user: account_user.auth_user
    map_attr('auth_user.email') <=> lambda account_user: account_user.auth_user.email
    """
    # there's no need to flip because of the order of arguments in the reduce function
    return partial(reduce, getattr, attr.split('.'))

def map_method(path, *args, **kwargs):
    """
    Returns a mapper function that runs the path method for each instance of
    the iterable collection.

    `map_method('accountuser_set.filter', is_deleted=False)` is equivalent to
    `lambda account: account.accountuser_set.filter(is_deleted=False)`
    """
    return lambda x: map_attr(path)(x)(*args, **kwargs)

def flatten(xs):
    """
    Flattens list of lists.

    >>> flatten([[1], [2], [3])
    >>> [1,2,3]
    >>> flatten([[10], [], [55])
    >>> [10, 55]
    """
    return reduce(iconcat, xs, [])

def zip_with_map(mapper, iterable):
    """
    Returns a collection of pairs where the first
    element correspond to the item in the iterable
    and the second is its mapped version, that is,
    the item when applied to the mapper function.

    >>> zip_with_map(lambda x: x**2, [1,2,3])
    >>> [(1,1), (2,4), (3,9)]
    """
    return zip(iterable, map(mapper, iterable))

def extend(*dicts):
    """
    Returns a dictionary that combines all dictionaries passed as arguments
    without mutating any of them.
    """
    def fold(acc, curr):
        acc.update(curr)
        return acc
    return reduce(fold, dicts, {})

def reduce_right(fold, xs, x0):
    """
    Right-associative fold of a structure.
    """
    return reduce(flip(fold), reversed(xs), x0)

identity = lambda x: x

class Identity:
    pass

_compose = lambda f, g: (
    f if g is Identity else
    lambda *args, **kwargs: f(g(*args, **kwargs))
)

def compose(*fns):
    """
    Simple function composition.
    """
    return reduce_right(_compose, fns, Identity)

def pipe(*fns):
    """
    Left-to-right composition, Unix style.
    """
    return reduce(flip(_compose), fns, Identity)



"""
exceptions module
"""

__author__ = 'Bruno Lange'
__email__ = 'blangeram@gmail.com'
__license__ = 'MIT'

class NonExhaustivePattern(Exception):
    """
    Thrown when pattern matching fails to find a matching predicate.
    """
    pass



import unittest
from functools import reduce, partial
import math
from exos import compose, pipe

class TestCompose(unittest.TestCase):
    def test_compose(self):
        f = lambda x: x*2
        h = lambda x: -x
        g = compose(f, h)
        self.assertEqual(g(4), f(h(4)))

        w = compose(h, f)
        self.assertEqual(w(4), h(f(4)))

        p = lambda x: x**2

        self.assertTrue(compose(p, g)(3) == p(f(h(3))) == 36)
        self.assertTrue(compose(h, p)(12) == h(p(12)) == -144)

    def test_general_composition(self):
        f = lambda x, y, bias=0: 2*x - math.sqrt(y) + bias
        g = lambda x: -x

        h = compose(g, f)
        self.assertEqual(h(5, 144), 2)

        hb = compose(g, partial(f, bias=100))
        self.assertEqual(hb(5, 144), -98)

        s = compose(str, h)
        self.assertEqual(s(12, 625), '1.0')

        digits = compose(lambda number: [int(digit) for digit in number.split('.')], s)
        self.assertEqual(digits(12, 625), [1, 0])

    def test_pipe(self):
        def f(x, y, bias=0):
            return 2*x - math.sqrt(y) + bias
        def g(x):
            return -x

        h = compose(g, f)
        p = pipe(f, g)
        self.assertTrue(h(5, 144) == p(5, 144))



import unittest
from functools import reduce
from exos import reduce_right

class TestReduceRight(unittest.TestCase):
    def test_commutative(self):
        sum1to100 = reduce(lambda acc, curr: acc+curr, range(1, 101), 0)
        sum100to1 = reduce_right(lambda curr, acc: acc+curr, range(1, 101), 0)
        self.assertTrue(sum1to100 == sum100to1 == 5050)



"""
stateful functions
"""

__author__ = 'Bruno Lange'
__email__ = 'blangeram@gmail.com'
__license__ = 'MIT'

def each(accept, iterable, *args, **kwargs):
    """
    Applies the accept function to each of the elements in the iterable
    collection.
    """
    if isinstance(accept, str):
        methods = [getattr(item, accept) for item in iterable]
        _ = [method(*args, **kwargs) for method in methods]
        return

    _ = (
        [accept(item) for item in iterable] if not kwargs.get('_unpack', False) else
        [accept(*item) for item in iterable]
    )

def ueach(accept, iterable, *args, **kwargs):
    """
    Unpacks elements in the collection before applying the accept function.
    """
    kwargs['_unpack'] = True
    each(accept, iterable, *args, **kwargs)

def print_each(xs, prefix=''):
    """
    Prints each element in the collection given by xs.
    """
    return ueach(
        lambda i, x: print('{}{}'.format(prefix, x).format(i=i)),
        enumerate(xs)
    )



"""
setup script for the exos module
"""

from setuptools import setup

__author__ = 'Bruno Lange'
__email__ = 'blangeram@gmail.com'
__license__ = 'MIT'

setup(
    name='exos',
    version='0.1.1',
    packages=['exos',],
    description='EXpressions Over Statements: additional functional tools in Python!',
    author='Bruno Lange',
    author_email='blangeram@gmail.com',
    url='https://github.com/brunolange/exos',
    python_requires='>=3',
)
